#!/usr/bin/env bash

setup_file(){
  load "$PWD/tests/integration/helpers/common"
  _common_setup
  _add_server
  _new_app
}

teardown_file(){
  _rm_app
  _rm_server
  _reset_recipe
}

setup(){
  load "$PWD/tests/integration/helpers/common"
  _common_setup
}

teardown(){
  # https://github.com/bats-core/bats-core/issues/383#issuecomment-738628888
  if [[ -z "${BATS_TEST_COMPLETED}" ]]; then
    _undeploy_app
  fi
}

@test "validate app argument" {
  run $ABRA app deploy
  assert_failure
  assert_output --partial 'no app provided'

  run $ABRA app deploy DOESNTEXIST
  assert_failure
  assert_output --partial 'cannot find app'
}

@test "bail if unstaged changes and no --chaos" {
  run bash -c "echo foo >> $ABRA_DIR/recipes/$TEST_RECIPE/foo"
  assert_success
  assert_exists "$ABRA_DIR/recipes/$TEST_RECIPE/foo"

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  assert_success
  assert_output --partial 'foo'

  run $ABRA app deploy "$TEST_APP_DOMAIN" --no-input
  assert_failure
  assert_output --partial 'locally unstaged changes'
  refute_output --partial 'chaos'

  run rm -rf "$ABRA_DIR/recipes/$TEST_RECIPE/foo"
  assert_not_exists "$ABRA_DIR/recipes/$TEST_RECIPE/foo"

  _checkout_recipe
}

# bats test_tags=slow
@test "do not bail if unstaged changes and --chaos" {
  run bash -c 'echo "unstaged changes" >> "$ABRA_DIR/recipes/$TEST_RECIPE/foo"'
  assert_success
  assert_exists "$ABRA_DIR/recipes/$TEST_RECIPE/foo"

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  assert_success
  assert_output --partial 'foo'

  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --chaos --no-input --no-converge-checks
  assert_success
  assert_output --partial 'chaos'

  _undeploy_app

  run rm -rf "$ABRA_DIR/recipes/$TEST_RECIPE/foo"
  assert_not_exists "$ABRA_DIR/recipes/$TEST_RECIPE/foo"
}

# bats test_tags=slow
@test "ensure recipe up to date if no --offline" {
  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" reset --hard HEAD~3
  assert_success

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  assert_output --partial 'behind 3'

  run $ABRA app deploy "$TEST_APP_DOMAIN" --no-input --no-converge-checks
  assert_success

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  refute_output --partial 'behind 3'

  _reset_recipe
  _undeploy_app
}

# bats test_tags=slow
@test "ensure recipe not up to date if --offline" {
  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" reset --hard HEAD~3
  assert_success

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  assert_output --partial 'behind 3'

  # NOTE(d1): need to use --chaos to force same commit
  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --chaos --offline
  assert_success

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  assert_output --partial 'behind 3'

  _undeploy_app
  _reset_recipe
}

# bats test_tags=slow
@test "deploy latest commit if no published versions and no --chaos" {
  latestCommit="$(git -C "$ABRA_DIR/recipes/$TEST_RECIPE" rev-parse --short HEAD)"

  _remove_tags

  # NOTE(d1): need to pass --offline to stop tags being pulled again
  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --offline
  assert_success
  assert_output --partial "$latestCommit"
  assert_output --partial 'using latest commit'
  refute_output --partial 'chaos'

  _undeploy_app
  _reset_tags
}

# bats test_tags=slow
@test "ensure same commit if --chaos" {
  latestCommit="$(git -C "$ABRA_DIR/recipes/$TEST_RECIPE" rev-parse --short HEAD)"

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" reset --hard HEAD~3
  assert_success

  run git -C "$ABRA_DIR/recipes/$TEST_RECIPE" status
  assert_output --partial 'behind 3'

  threeCommitsBack="$(git -C "$ABRA_DIR/recipes/$TEST_RECIPE" rev-parse --short HEAD)"

  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --chaos
  assert_success
  refute_output --partial "$latestCommit"
  assert_output --partial "$threeCommitsBack"
  assert_output --partial 'chaos'

  _undeploy_app
  _reset_recipe
}

# bats test_tags=slow
@test "retrieve recipe if missing" {
  run rm -rf "$ABRA_DIR/recipes/$TEST_RECIPE"
  assert_success
  assert_not_exists "$ABRA_DIR/recipes/$TEST_RECIPE"

  run $ABRA app deploy "$TEST_APP_DOMAIN" --no-input --no-converge-checks
  assert_success

  _undeploy_app

  assert_exists "$ABRA_DIR/recipes/$TEST_RECIPE"
}

@test "no deploy if lint error" {
  run sed -i '/traefik.enable=.*/d' "$ABRA_DIR/recipes/$TEST_RECIPE/compose.yml"
  assert_success

  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --chaos
  assert_failure
  assert_output --partial 'failed lint checks'

  _checkout_recipe
}

# bats test_tags=slow
@test "error if already deployed and no --force/--chaos" {
  _deploy_app

  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks
  assert_failure
  assert_output --partial 'already deployed'

  _undeploy_app
}

# bats test_tags=slow
@test "re-deploy deployed app if --force/--chaos" {
  _deploy_app

  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --force
  assert_success
  assert_output --partial 'already deployed but continuing'
  assert_output --partial '--force'

  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --chaos
  assert_success
  assert_output --partial 'already deployed but continuing'
  assert_output --partial '--chaos'

  _undeploy_app
}

# bats test_tags=slow
@test "deploy latest version from catalogue if no --chaos" {
  latestVersion=$(jq -r '.gitea.versions[-1] | keys[0]' < "$ABRA_DIR/catalogue/recipes.json")
  refute [ -z "$latestVersion" ];

  run $ABRA app new gitea \
    --no-input \
    --server "$TEST_SERVER" \
    --domain "gitea.$TEST_SERVER" \
    --secrets
  assert_success
  assert_exists "$ABRA_DIR/servers/$TEST_SERVER/gitea.$TEST_SERVER.env"

  run $ABRA app deploy "gitea.$TEST_SERVER" --no-input --no-converge-checks
  assert_success
  assert_output --partial "$latestVersion"

  run $ABRA app undeploy "gitea.$TEST_SERVER" --no-input
  assert_success

  run $ABRA app secret remove "gitea.$TEST_SERVER" --all --no-input
  assert_success

  # NOTE(d1): to let the stack come down before nuking volumes
  sleep 5

  run $ABRA app volume remove "gitea.$TEST_SERVER" --no-input
  assert_success

  run $ABRA app remove "gitea.$TEST_SERVER" --no-input
  assert_success
  assert_not_exists "$ABRA_DIR/servers/$TEST_SERVER/gitea.$TEST_SERVER.env"
}

# bats test_tags=slow
@test "skip domain check if missing DOMAIN=" {
  run cp "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env" "$BATS_TMPDIR/$TEST_APP_DOMAIN.env"
  assert_success
  assert_exists "$BATS_TMPDIR/$TEST_APP_DOMAIN.env"

  run grep -q "DOMAIN=" "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_success

  run sed -i '/DOMAIN=.*/d' "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_success

  run grep -q "DOMAIN=" "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_failure

  run $ABRA app deploy "$TEST_APP_DOMAIN" --no-input --no-converge-checks
  assert_success
  assert_output --partial 'no DOMAIN=... configured for app'

  run $ABRA app undeploy "$TEST_APP_DOMAIN" --no-input
  assert_success

  run mv "$BATS_TMPDIR/$TEST_APP_DOMAIN.env" "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_success
}

@test "ensure domain is checked" {
  appDomain="custom-html.DOESNTEXIST"

  run $ABRA app new custom-html \
    --no-input \
    --server "$TEST_SERVER" \
    --domain "$appDomain"
  assert_success
  assert_exists "$ABRA_DIR/servers/$TEST_SERVER/$appDomain.env"

  run $ABRA app deploy "$appDomain" --no-input
  assert_failure
  assert_output --partial 'no such host'
}

# bats test_tags=slow
@test "skip domain check when requested" {
  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --no-domain-checks
  assert_success
  assert_output --partial 'skipping domain checks as requested'

  _undeploy_app
}

@test "error if specific version does not exist" {
  run $ABRA app deploy "$TEST_APP_DOMAIN" DOESNTEXIST --no-input --no-converge-checks
  assert_failure
  assert_output --partial "doesn't seem to have version DOESNTEXIST available"
}

# bats test_tags=slow
@test "deploy specific version" {
  run $ABRA app deploy "$TEST_APP_DOMAIN" "0.2.0+1.21.0" --no-input
  assert_success
  assert_output --partial "0.2.0+1.21.0"

  _undeploy_app
}

@test "bail out if specific version and chaos" {
  run $ABRA app deploy "$TEST_APP_DOMAIN" "0.2.0+1.21.0" \
    --chaos --no-input --no-converge-checks
  assert_failure
  assert_output --partial 'cannot use'
}

# bats test_tags=slow
@test "COMPOSE_FILE with \$COMPOSE_FILE override works" {
  _reset_recipe

  run sed -i 's/#COMPOSE_FILE="$COMPOSE_FILE:compose.extra_env.yml"/COMPOSE_FILE="$COMPOSE_FILE:compose.extra_env.yml"/g' \
    "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_success

  # NOTE(d1): --chaos used to bypass versions and access compose.extra_env.yml
  run $ABRA app deploy "$TEST_APP_DOMAIN" \
    --no-input --no-converge-checks --chaos
  assert_success
  assert_output --partial "compose.yml"
  assert_output --partial "compose.extra_env.yml"

  _undeploy_app
  _reset_app
}

@test "error if no secrets generated" {
  run sed -i 's/COMPOSE_FILE="compose.yml"/COMPOSE_FILE="compose.yml:compose.extra_secret.yml"/g' \
    "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_success

  run sed -i 's/#SECRET_EXTRA_PASS_VERSION=v1/SECRET_EXTRA_PASS_VERSION=v1/g' \
    "$ABRA_DIR/servers/$TEST_SERVER/$TEST_APP_DOMAIN.env"
  assert_success

  run $ABRA app deploy "$TEST_APP_DOMAIN" --no-input --no-converge-checks
  assert_failure
  assert_output --partial 'unable to deploy, secrets not generated'

  _reset_app
}

@test "recipe config comments not present in values" {
  run $ABRA app deploy "$TEST_APP_DOMAIN" --no-input
  assert_success

  run $ABRA app run "$TEST_APP_DOMAIN" app env
  assert_success
  refute_output --partial 'should be removed'
}
